#
# Copyright 2017 CHLSoftware
#
# A set of utilities for connecting to a DocMoto server.
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import urllib2
from xml.dom.minidom import parse, parseString
import socket
import ssl


#Override urllib2 function to overcome issue with Negotiate
original_function = urllib2.parse_keqv_list

 
def wrapper(l):
    
    out = []
    for item in l:
        if not "Negotiate" in str(item):
            out.append(item)
    
    l = out
    return original_function(l)
 
urllib2.parse_keqv_list = wrapper


#Create our own opener because urllib2 does not retain authorization
#data, so sends everything twice. Which apart from being a bit wasteful also
#fails with a broken pipe error for a moderately large file
class CHLHTTPDigestAuthHandler(urllib2.HTTPDigestAuthHandler):

    auth_val = None
    handler_order = 100
    retried = False
    realm = None
    nonce = None
    qop = None
    algorithm = None
    opaque = None
    
    #Add the authorize header
    def http_request(self,request):
        
        if(CHLHTTPDigestAuthHandler.retried):
            #refresh authentication details
            auth = CHLHTTPDigestAuthHandler.get_authorization(self, request)
            CHLHTTPDigestAuthHandler.auth_val = 'Digest %s' % auth
            request.add_header("Authorization",CHLHTTPDigestAuthHandler.auth_val)
            
        
        return request
    
    https_request = http_request
    
    #Overrides urllib2 method
    #Only to be used the once when plain request made.
    #The response needs recalculating each time  as based on method (PUT, PROPFIND URI)
    def retry_http_digest_auth(self, req, auth):
        token, challenge = auth.split(' ', 1)
        chal = urllib2.parse_keqv_list(urllib2.parse_http_list(challenge))
        auth = self.get_authorization(req, chal)

        if auth:
            CHLHTTPDigestAuthHandler.auth_val = 'Digest %s' % auth
            
            if CHLHTTPDigestAuthHandler.retried:
                return None
            
            if req.headers.get(self.auth_header, None) == CHLHTTPDigestAuthHandler.auth_val:
                return None
            
            req.add_unredirected_header(self.auth_header, CHLHTTPDigestAuthHandler.auth_val)
            resp = self.parent.open(req, timeout=req.timeout)
            CHLHTTPDigestAuthHandler.retried = True
            
            return resp
        
    #Overrides urllib2 method. Necessary to recalculate response value
    #for each request
    def get_authorization(self, req, chal=None):
        try:
            if (chal):
                CHLHTTPDigestAuthHandler.realm = chal['realm']
                CHLHTTPDigestAuthHandler.nonce = chal['nonce']
                CHLHTTPDigestAuthHandler.qop = chal.get('qop')
                CHLHTTPDigestAuthHandler.algorithm = chal.get('algorithm', 'MD5')
                # mod_digest doesn't send an opaque, even though it isn't
                # supposed to be optional
                CHLHTTPDigestAuthHandler.opaque = chal.get('opaque', None)
        except KeyError:
            return None

        H, KD = self.get_algorithm_impls(CHLHTTPDigestAuthHandler.algorithm)
        if H is None:
            return None

        user, pw = self.passwd.find_user_password(CHLHTTPDigestAuthHandler.realm, req.get_full_url())
        if user is None:
            return None

        # not implemented yet
        if req.has_data():
            entdig = self.get_entity_digest(req.get_data(), chal)
        else:
            entdig = None

        A1 = "%s:%s:%s" % (user, CHLHTTPDigestAuthHandler.realm, pw)
        A2 = "%s:%s" % (req.get_method(),
                        # XXX selector: what about proxies and full urls
                        req.get_selector())
        if CHLHTTPDigestAuthHandler.qop == 'auth':
            if CHLHTTPDigestAuthHandler.nonce == self.last_nonce:
                self.nonce_count += 1
            else:
                self.nonce_count = 1
                self.last_nonce = nonce

            ncvalue = '%08x' % self.nonce_count
            cnonce = self.get_cnonce(nonce)
            noncebit = "%s:%s:%s:%s:%s" % (CHLHTTPDigestAuthHandler.nonce, ncvalue, cnonce, CHLHTTPDigestAuthHandler.qop, H(A2))
            respdig = KD(H(A1), noncebit)
        elif CHLHTTPDigestAuthHandler.qop is None:
            respdig = KD(H(A1), "%s:%s" % (CHLHTTPDigestAuthHandler.nonce, H(A2)))
        else:

            raise URLError("qop '%s' is not supported." % CHLHTTPDigestAuthHandler.qop)

        base = 'username="%s", realm="%s", nonce="%s", uri="%s", ' \
               'response="%s"' % (user, CHLHTTPDigestAuthHandler.realm, CHLHTTPDigestAuthHandler.nonce, req.get_selector(),
                                  respdig)           
                        
        
        if CHLHTTPDigestAuthHandler.opaque:
            base += ', opaque="%s"' % CHLHTTPDigestAuthHandler.opaque
        if entdig:
            base += ', digest="%s"' % entdig
        base += ', algorithm="%s"' % CHLHTTPDigestAuthHandler.algorithm
        if CHLHTTPDigestAuthHandler.qop:
            base += ', qop=auth, nc=%s, cnonce="%s"' % (ncvalue, cnonce)  
            
        return base
    
    def get_entity_digest(self, data, chal=None):
        # Not implemented yet
        return None

# host in form https://xxx.xxx.xxx.xxx:3984
def secureinitialize(userName,userPassword,host):
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    passwdmngr = urllib2.HTTPPasswordMgrWithDefaultRealm()
    passwdmngr.add_password('DocMotoRealm', host, userName, userPassword)
    auth_handler = CHLHTTPDigestAuthHandler(passwdmngr)
    opener = urllib2.build_opener(urllib2.HTTPSHandler(context=ctx), auth_handler)
    urllib2.install_opener(opener)
    req = urllib2.Request(host + "/")
    req.get_method = lambda: 'OPTIONS'
    return doRequest(req)

# host in form http://xxx.xxx.xxx.xxx:3983
# DO NOT USE in production, only use secure version
def initialize(userName,userPassword,host):

    passwdmngr = urllib2.HTTPPasswordMgrWithDefaultRealm()
    passwdmngr.add_password('DocMotoRealm', host, userName, userPassword)
    auth_handler = CHLHTTPDigestAuthHandler(passwdmngr)
    opener = urllib2.build_opener(auth_handler)
    urllib2.install_opener(opener)
    req = urllib2.Request(host + "/")
    req.get_method = lambda: 'OPTIONS'
    return doRequest(req)
   
def doRequest(request):
    status = False
    code = 0
    responseString = ''

    try:    
        response = urllib2.urlopen(request,None,120)
      
    except urllib2.HTTPError, e:

        status = False
        code = e.code
        responseString = str(e.strerror)

        
    except urllib2.URLError, e:
        status = False
        code = e.reason[0]
        responseString = str(e.reason[1])

        
    except socket.timeout:

        status = False
        code = 0
        responseString = 'Socket timed out'
      
    except IOError, e:

        status = False
        code = e.errno
        responseString = str(e.strerror)
        
    except:
        status = False
        code = 0
        responseString = 'Unexpected error'
 
    else:
        status = True
        code = response.getcode()
        responseString = str(response.read())
        
    finally:
        return {'returnOK':status,'code':code,'responseString':responseString}      
        

def createCollection(urlDoc):
    req = urllib2.Request(urlDoc)
    req.get_method = lambda: 'MKCOL'
    return doRequest(req)


def createPut(urlDoc,content):
    
    req = urllib2.Request(urlDoc, content)
    req.get_method = lambda: 'PUT'
    req.add_header("Content-Type","application/octet-stream")
    req.add_header("Content-Length",len(content))
    #req.add_data = str(content)
    #req.data = content

    return doRequest(req)


def createPropPatch(urlDoc,content):
    req = urllib2.Request(urlDoc, data=content)
    req.get_method = lambda: 'PROPPATCH'
    return doRequest(req)


def createVersionControl(urlDoc,fromRevision=''):
    req = urllib2.Request(urlDoc)
    req.get_method = lambda: 'VERSION-CONTROL'
    
    if fromRevision :
        req.add_data(fromRevision)

    return doRequest(req)

  
def createPropFind(urlDoc,depth):
    req = urllib2.Request(urlDoc)    
    req.get_method = lambda: 'PROPFIND'    
    req.add_header("Depth", depth)
    return doRequest(req)
    
def createDelete(urlDoc):
    req = urllib2.Request(urlDoc)
    req.get_method = lambda: 'DELETE'
    return doRequest(req)

#propNamespace will be something like DAV:, DOCMOTO: or DMUSER: Don't forget the :
def propGetSingleProperty(urlDoc,propName,propNamespace):
    myRet = createPropFind(urlDoc, 1)
    responseXMLString = myRet['responseString']
    return getValueFromXML(responseXMLString,propName,propNamespace)


#Used to get property values from raw returned XML. Returns empty string if problem
def getValueFromXML(responseXMLString,propName,propNamespace):
    if not responseXMLString:
        return ""
    try:
        parseXMLdoc = parseString(responseXMLString)
        xmlTag = parseXMLdoc.getElementsByTagNameNS(propNamespace, propName)[0]
    
        if(xmlTag.hasChildNodes()):
            if xmlTag.childNodes[0].nodeValue == None:
                return ""
            else:
                return str( xmlTag.childNodes[0].nodeValue)

    except:
        return ""

#Returns true if exists, false otherwise, also lets you know if a folder or not
def checkExists(urlDoc):
    myRet = createPropFind(urlDoc, 1)
    
    if myRet['code'] != 207:
        return {'exists':False, 'isCollection':False}
    
    if isCollection(myRet['responseString']):
        return {'exists':True, 'isCollection':True}
    else:
        return {'exists':True, 'isCollection':False}        
    
 
def isCollection(responseXMLString):
   
    try:
        parseXMLdoc = parseString(responseXMLString)
        xmlTag = parseXMLdoc.getElementsByTagNameNS("DAV:", 'resourcetype')[0]
           
        if(xmlTag.hasChildNodes()):
     
            if str(xmlTag.childNodes[0].nodeName).lower().find("collection") > 0:
                return True
            else:
                return False
        else:
            return False
        
    except:
        return False